/*
 * main.c - args, configuration, logging
 *
 * MetaServerII relay
 */
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/param.h>
#include <errno.h>
#include "relay.h"


/*
 * Globals
 */
int verbose = FALSE;		/* if TRUE, print extra info on terminal */
char *prog = NULL;		/* our program name */
PORT *ports = NULL;		/* list of ports to listen on */
int port_count = 0;		/* #of ports in list */
time_t uptime = 0;		/* when we started */



/*
 * Private variables
 */
static FILE *logfp = NULL;	/* log file */
static int do_log = TRUE;	/* set from config file */
static int truncate = 0;	/* truncate log file on startup? */

PORT_DEF port_defs[] = {
    { DI_BUFCP, "bufcp" },
};
#define MAX_PORT_DEF    (sizeof(port_defs) / sizeof(PORT_DEF))


/*
 * Configuration stuff
 */
typedef enum {
    STMT_BOOL, STMT_INT, STMT_STRING, STMT_SERVER, STMT_PORT
} STMT_KIND;
struct keyword_t {
    STMT_KIND kind;
    char *keyword;
    int *var;
} keywords[] = {
    { STMT_BOOL,	"do_logging",	&do_log },
    { STMT_PORT,	"port",		NULL },
};
#define MAX_KEYWORDS	(sizeof(keywords) / sizeof(struct keyword_t))


/*
 * print a message to the log
 *
 * (I'd use varargs, but vfprintf() doesn't exist everywhere, so why bother?)
 */
void
log_msg(format, arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7)
char *format;
int arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7;
{
    char buf[256];

    if (!do_log) return;
    if (logfp == NULL) {
	fprintf(stderr, "WARNING: attempt to write message to closed log\n");
	return;
    }

    /* date is in hex to keep it compact */
    sprintf(buf, "%.8lx: ", time(0));
    sprintf(buf+10, format, arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7);
    if (*(buf + strlen(buf)-1) != '\n')
	strcat(buf, "\n");

    /* since there should only be one metaserver process, we probably don't */
    /* need to seek to the end every time (but it allows truncation...) */
    fseek(logfp, (off_t) 0, 2);
    fprintf(logfp, buf);
}


/* add an entry to the "port" list */
static void
add_port(port, type, rhost, rport, stale)
char *port, *type, *rhost, *rport, *stale;
{
    DISPLAY kind;
    int portnum = atoi(port);
    int rportnum = atoi(rport);
    int i;

    /* see if portnum is valid and port kind is known */
    if (portnum < 1024) {
	fprintf(stderr, "WARNING: port %d (%s) reserved; stmt ignored\n",
		portnum, port);
	return;
    }
    if (!rportnum) {
	fprintf(stderr, "WARNING: remote port %d (%s) invalid; stmt ignored\n",
		rportnum, rport);
	return;
    }
    for (i = 0; i < MAX_PORT_DEF; i++) {
	if (!strcmp(type, port_defs[i].port_name)) {
	    kind = port_defs[i].port_kind;
	    break;
	}
    }
    if (i == MAX_PORT_DEF) {
	fprintf(stderr, "WARNING: unknown port kind '%s'\n", type);
	return;
    }

    /* create a new entry, and add it to the list */
    if (ports == NULL) {
	ports = (PORT *) malloc(sizeof(PORT));
    } else {
	ports = (PORT *) realloc(ports,
		sizeof(PORT) * (port_count+1));
    }
    if (ports == NULL) {
	fprintf(stderr, "ERROR: malloc/realloc failure (ports)\n");
	exit(1);
    }
    ports[port_count].port = portnum;
    ports[port_count].kind = kind;
    strcpy(ports[port_count].rhost, rhost);
    if ((ports[port_count].raddr = lookup(rhost)) == -1) {
	fprintf(stderr, "WARNING: unable to find host '%s'\n", rhost);
	return;
    }

    ports[port_count].rport = atoi(port);
    ports[port_count].stale = atoi(stale);
#ifdef DEBUG
    printf("port=%d, kind=%d, rhost='%s' (0x%.8lx), rport=%d, stale=%d\n",
	ports[port_count].port, ports[port_count].kind,
	ports[port_count].rhost, ports[port_count].raddr,
	ports[port_count].rport, ports[port_count].stale);
#endif
    port_count++;
}

/* find the next token */
static char *
next_token(str)
char *str;
{
    /* skip leading whitespace */
    while (*str && (*str == ' ' || *str == '\t' || *str == '\n'))
	str++;
    if (*str)
	return (str);
    else
	return (NULL);
}


/*
 * read the configuration in (may be called from a signal handler)
 */
static void
load_config()
{
    FILE *fp;
    char buf[128], *cp, *cp2, *cp3, *cp4, *cp5;
    int i, len, line;

    /* clear out any existing configuration stuff */
    if (port_count) {
	free(ports);
	port_count = 0;
    }

    if ((fp = fopen(CONFIGFILE, "r")) == NULL) {
	perror("ERROR: can't open config file");
	exit(1);
    }

    line = 0;
    while (1) {
	fgets(buf, 128, fp);
	line++;
	if (feof(fp)) break;
	if (ferror(fp)) {
	    perror("ERROR: error reading config file");
	    exit(1);
	}
	if (buf[strlen(buf)-1] == '\n')
	    buf[strlen(buf)-1] = '\0';

	/* skip comments and blank lines */
	cp = next_token(buf);
	if (cp == NULL || *cp == '#')
	    continue;

	/* find a token in "keywords" which matches */
	for (i = 0; i < MAX_KEYWORDS; i++) {
	    len = strlen(keywords[i].keyword);	/* use table for speed? */
	    if (!strncmp(cp, keywords[i].keyword, len)) {
		cp2 = next_token(cp+len);
		if (cp2 != NULL && *cp2 == ':') {
		    break;
		}
	    }
	}
	if (i == MAX_KEYWORDS) {
	    fprintf(stderr, "WARNING: unknown statement (line %d): %s\n",
		line, buf);
	    continue;
	}

	/* parse the right-hand side with strtok (unless it's a string) */
	if (keywords[i].kind == STMT_STRING || keywords[i].kind == STMT_SERVER){
	    cp = next_token(cp2+1);
	} else if ((cp = strtok(cp2+1, " \t")) == NULL) {
	    fprintf(stderr, "WARNING: missing rhs (line %d); ignored\n", line);
	    continue;
	}

	switch (keywords[i].kind) {
	case STMT_BOOL:		/* boolean on/off */
	    if (!strcasecmp(cp, "on"))
		*(keywords[i].var) = 1;
	    else if (!strcasecmp(cp, "off"))
		*(keywords[i].var) = 0;
	    else
		fprintf(stderr,
		    "WARNING: boolean expr must be 'on' or 'off' (line %d)\n",
		    line);
	    break;
	case STMT_INT:		/* integer value */
	    *((int *) keywords[i].var) = atoi(cp);
	    break;
	case STMT_STRING:	/* string; dest should be allocated */
	    if (keywords[i].var == NULL) {
		fprintf(stderr, "WARNING: no storage\n");
	    } else {
		strcpy((char *) keywords[i].var, cp);
	    }
	    break;
	case STMT_PORT:		/* "port" line (yeah, it's ugly) */
	    if ((cp2 = strtok(NULL, " \t")) == NULL) {
		fprintf(stderr, "WARNING: port missing port kind; ignored\n");
		break;
	    }
	    if ((cp3 = strtok(NULL, "\"\t")) == NULL) {
		fprintf(stderr, "WARNING: port missing description; ignored\n");
		break;
	    }
	    if ((cp4 = strtok(NULL, "\"\t")) == NULL) {
		fprintf(stderr, "WARNING: port missing description; ignored\n");
		break;
	    }
	    cp5 = strtok(NULL, " \t");
	    add_port(cp, cp2, cp3, cp4, cp5);
	    break;
	default:
	    /* "shouldn't happen" */
	    fprintf(stderr, "Internal error: bad keyword kind %d\n",
		keywords[i].kind);
	}
    }

    /* sanity check: make sure we have something to do! */
    if (!port_count) {
	fprintf(stderr, "ERROR: no ports to listen on!\n");
	exit(1);
    }
}


/*
 * open the log file and print a friendly message
 */
static void
open_log()
{
    time_t now;
    char *ostr;

    if (!do_log) return;

    /* prepare the log file */
    if (!truncate)
	ostr = "a";	/* just append */
    else
	ostr = "w";	/* truncate */

    if ((logfp = fopen(LOGFILE, ostr)) == NULL) {
	perror("ERROR: can't open log file for writing");
	exit(1);
    } else {
	setvbuf(logfp, NULL, _IONBF, 0);   /* no buffering */

	log_msg("\n");
	now = time(0);
	log_msg("** MetaServerII relay v%s started on %s",VERSION,ctime(&now));
    }
}


/* fatal signal received */
static int
sig_death(sig, code, scp, addr)
int sig, code;
struct sigcontext *scp;
char *addr;
{
    if (verbose) printf("%s: killed by signal %d\n", prog, sig);
    log_msg("exit: killed by signal %d", sig);

#ifdef ABORT
#ifndef PROF	/* always exit cleanly if we're doing profiling */
    log_msg("sig = %d, code = %d, scp = 0x%.8lx, addr = 0x%.8lx\n",
	sig, code, scp, addr);
    log_msg("pc = 0x%.8lx\n", scp->sc_pc);
    log_msg("Dumping core...\n");
    abort();
#endif
#endif /*ABORT*/

    exit(2);
}


/* reconfigure signal receieved */
static int
sig_reconfig(sig)
int sig;
{
    if (verbose) printf("%s: received reconfigure signal %d\n", prog, sig);
    /*log_msg("rereading configuration file (signal %d)", sig);*/
    log_msg("ignoring config signal %d", sig);

    /*load_config();*/
}


/*
 * initialize signal handling stuff
 */
static void
init_signals()
{
    SIGNAL(SIGINT, sig_death);
    SIGNAL(SIGQUIT, sig_death);
    SIGNAL(SIGTERM, sig_death);
    SIGNAL(SIGHUP, sig_reconfig);
    SIGNAL(SIGPIPE, SIG_IGN);

    SIGNAL(SIGALRM, wakeup);
}


/*
 * Process args
 */
int
main(argc, argv)
int argc;
char *argv[];
{
    extern char *optarg;
    extern int optind;
    int c;

    /* parse args */
    while ((c = getopt(argc, argv, "vt")) != EOF)
	switch (c) {
	case 'v':
	    verbose++;
	    break;
	case 't':
	    truncate++;
	    break;
	case '?':
	default:
	    fprintf(stderr, "Netrek MetaServerII relay v%s\n", VERSION);
	    fprintf(stderr, "usage: relayII [-v] [-t]\n\n");
	    exit(2);
	}

    if ((prog = strrchr(argv[0], '/')) == NULL)
	prog = argv[0];
    else
	prog++;

    /* read the configuration file */
    load_config();

    /* initialize internal data structures */
    init_connections();

    /* open the log and print a friendly message */
    if (do_log) open_log();

    /* init the signal handling stuff */
    init_signals();

    uptime = time(0);

    /* nothing has failed yet, so go wait for connections */
    main_loop();

    exit(0);
    /*NOTREACHED*/
}

