/*
 * scan.c - scan the various servers
 *
 * MetaServerII relay
 */
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include "relay.h"

/*
 * Server connection globals
 */
int busy = -1,
    last_busy = 0,
    busy_sock = -1,
    busy_time = 0;

USER *users;

/*
 * Private variables
 */
static fd_set readfd_set, writefd_set;
static int maxfd;
static /*volatile*/ int coffee;	/* wake up and smell the */


/*
 * ---------------------------------------------------------------------------
 *	initialization/setup
 * ---------------------------------------------------------------------------
 */

/*
 * Find a remote host given on the port list.
 *
 * Converts the name from ASCII IP or DNS format to an addr.  (Used by
 * add_port.)
 */
long
lookup(host)
char *host;
{
    struct hostent *hp;
    long addr;

    if ((addr = inet_addr(host)) == -1) {
	if (verbose) printf("--- Looking up %s\n", host);
        if ((hp = gethostbyname(host)) == NULL) {
            fprintf(stderr, "WARNING: unknown host %s on port list\n", host);
	    /*log_msg("unknown host %s on server list", host);   (not open!) */
            return(-1);
        } else {
            addr = *(long *) hp->h_addr;
        }
    }

    return (addr);
}

/*
 * Wake up and smell the users (SIGALRM handler).
 */
int
wakeup()
{
    coffee = TRUE;
}


/*
 * Do some silly setup stuff.
 */
void
init_connections()
{
    int i;

    maxfd = getdtablesize();
    if (maxfd > FD_SETSIZE)
	maxfd = FD_SETSIZE;
    if (verbose) printf("%s: maximum file descriptor is %d\n", prog, maxfd);

    /* one struct for every fd */
    if ((users = (USER *) malloc(maxfd * sizeof(USER))) == NULL) {
	fprintf(stderr, "malloc failure (users)\n");
	log_msg("exit: malloc failure in init_connections\n");
	exit(1);
    }
    bzero(users, maxfd * sizeof(USER));		/* force start_time, buf to 0 */

    /* init MS-II connection stuff for ports */
    for (i = 0; i < port_count; i++) {
	ports[i].data = NULL;
	ports[i].cstate = CS_CLOSED;
	ports[i].why_dead = WD_UNKNOWN;
	ports[i].sock = -1;
    }

}


/*
 * ---------------------------------------------------------------------------
 *	low-level connection stuff
 * ---------------------------------------------------------------------------
 */

#ifdef DEBUG
/* debugging */
printPeerInfo(sock)
int sock;
{
    struct sockaddr_in addr;
    int len;

    len = sizeof(addr);
    if (getsockname(sock, &addr, &len) < 0) {
/*      perror("printUdpInfo: getsockname");*/
        return;
    }
    log_msg("+ LOCAL: addr=0x%x, family=%d, port=%d\n", addr.sin_addr.s_addr,
        addr.sin_family, ntohs(addr.sin_port));

    if (getpeername(sock, &addr, &len) < 0) {
/*      perror("printUdpInfo: getpeername");*/
        return;
    }
    log_msg("+ PEER : addr=0x%x, family=%d, port=%d\n", addr.sin_addr.s_addr,
        addr.sin_family, ntohs(addr.sin_port));
}
#endif


/*
/*
 * write a bunch of data
 */
/*static*/ int
write_data(fd, buf, count)
int fd;
register char *buf;
register int count;
{
    register long n;
    int orig = count;

    for (; count ; buf += n, count -= n) {
	if ((n = write(fd, buf, count)) < 0) {
	    perror("write_data");
	    if (errno == EPIPE)
		log_msg("%.2d unable to connect (EPIPE)", fd);
	    else
		log_msg("%.2d write error %d", fd, errno);
	    return (-1);
	}
    }
    return (orig);
}


/*
 * We have a new connection from a user.  Get the requested information and
 * add the user to the list of fds we watch for writeability.  Don't actually
 * transmit any data yet.
 *
 * If it looks like a particular host is flooding us (> 3 calls in 10 seconds),
 * ignore him.  This is a rudimentary scheme which gets reset if somebody else
 * calls, so a flooder could just call from two different hosts at the same
 * time.
 *
 * This does NOT reset the flood_time on every rejected call, so after 10
 * seconds all is (mostly) forgiven.  This avoids the problem of somebody
 * who likes to bang his head against the wall.
 */
void
new_user(idx, sock)
int idx, sock;
{
    struct sockaddr_in addr, from;
    struct hostent *hp;
    struct linger ling;
    char *host;
    int csock, port;
    int length;
    static time_t flood_start = 0;
    static int flood_count = 0, flood_addr = 0;
    time_t now;

#ifdef DEBUG
    printf("Got client connection!\n");
#endif
    length = sizeof(addr);
    if ((csock = accept(sock, &addr, &length)) < 0) {
	/* this can happen if we get too many connections, or file tab ovfl */
	log_msg("user accept failed, err = %d\n", errno);
	/*perror("WARNING: accept failed");*/
	return;
    }

    /* check for floods */
    if (addr.sin_addr.s_addr == flood_addr) {
	/* at least two in a row */
	now = time(0);
	if (!flood_start) flood_start = now;

	if (now > flood_start + FLOOD_TIME) {
	    /* he's moving slowly, but keep an eye on him */
	    flood_start = now;
	    flood_count = -1;
	}
	if (++flood_count >= MAX_CALLS) {
	    /* he's flooding, knock him */
	    close(csock);
	    log_msg("FLOOD 0x%.8lx", addr.sin_addr.s_addr);
	    return;
	}
    } else {
	/* flood_start is set to 0 to avoid extra time(0) call */
	/* (slightly silly, but we're really only worried about psychos) */
	flood_addr = addr.sin_addr.s_addr;
	flood_start = flood_count = 0;
    }


#ifdef DEBUG
    length = sizeof(ling);
    if (getsockopt(csock, SOL_SOCKET, SO_LINGER, &ling, &length) < 0)
	perror("getsockopt");
    printf("(len = %d) linger on=%d, time=%d\n", length,
	    ling.l_onoff, ling.l_linger);
#endif
    ling.l_onoff = 1;	/* turn it on */
    ling.l_linger = /*3*/  10;
    length = sizeof(ling);
    if (setsockopt(csock, SOL_SOCKET, SO_LINGER, &ling, length) < 0)
	perror("setsockopt");

    /* figure out who we're connected to for logging purposes */
    length = sizeof(from);
    if (getpeername(csock, (struct sockaddr *) &from, &length) < 0) {
	host = "unknown";
	port = 0;
    } else {
	hp = gethostbyaddr((char *)&from.sin_addr.s_addr,sizeof(long),AF_INET);
	if (hp != NULL)
	    host = hp->h_name;
	else
	    host = inet_ntoa(from.sin_addr);
	port = ntohs(from.sin_port);
    }
    log_msg("Conn to svc %d from %s (%d)  (s=%d)", idx, host, port, csock);

    switch (ports[idx].kind) {
    case DI_BUFCP:
	svc_bufcp(csock, idx);
	break;
    default:
	log_msg("Internal ERROR: unknown service %d", ports[idx].kind);
	close(csock);
	return;
    }

    users[csock].start_time = time(0);
    FD_SET(csock, &writefd_set);
}


/*
 * Handle a user connection.
 *
 * Calls close_user() after all data is written.
 */
/*static*/ int
handle_user(fd)
{
    int cc;

    if ((cc = write(fd, users[fd].data->buf + users[fd].pos,
	users[fd].data->data_size - users[fd].pos)) < 0){
#ifdef DEBUG
	perror("write in handle_user");
#endif
	log_msg("write error in handle_user = %d", errno);
	close_user(fd);
	return (-1);
    }
#ifdef DEBUG2
    printf("wrote %d bytes to user\n", cc);
#endif
    users[fd].pos += cc;
    /*users[fd].data_size -= cc;*/

    if (users[fd].pos == users[fd].data->data_size) {
	/* all done! */
	close_user(fd);
	return (1);
    }

    return (0);
}


/*
 * Close a user connection.
 */
/*static*/ int
close_user(fd)
{
    if (users[fd].pos != users[fd].data->data_size) {
	log_msg("hosed user on fd %d after %d seconds\n", fd,
	    time(0) - users[fd].start_time);
    }
    close(fd);			/* this can linger */
    FD_CLR(fd, &writefd_set);
    /*free(users[fd].buf);*/
    users[fd].data = NULL;
    users[fd].start_time = 0;	/* mark as inactive */

    return (0);
}


#ifdef FUBAR
/*
 * Open a new connection to a server.  Doesn't actually do anything; just
 * opens it non-blocking.
 *
 * Set busy before calling.  Sets busy_sock.
 */
/*static*/ int
open_server()
{
    struct sockaddr_in addr;
    int sock;
    int flags;

#ifdef DEBUG2
    { int i;
    for (i = 0; i < server_count; i++) {
	printf("%d: %s\n", i, servers[i].hostname);
    }}
#endif

    if (servers[busy].cstate != CS_CLOSED)
	fprintf(stderr, "Internal WARNING: cstate != CLOSED\n");
    if (verbose)
	printf("%s: Opening connection to #%d: %s (%d)\n", prog, busy,
	    servers[busy].hostname, servers[busy].port -
	    ((servers[busy].pstate == PS_PORTM) ? 1 : 0));

    /* establish TCP connection to server */
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
#ifdef DEBUG
	perror("socket (server connection)");
#endif
	log_msg("socket() failed, err=%d\n", errno);
	return (-1);
    }

    if ((flags = fcntl(sock, F_GETFL, 0)) < 0) {
	/*perror("unable to get flags for NB");*/
	log_msg("fcntl(get) failed, err=%d\n", errno);
	return(-1);
    }
    if (fcntl(sock, F_SETFL, flags|O_NDELAY) < 0) {
	/*perror("unable to set NB mode for connect");*/
	log_msg("fcntl(set) failed, err=%d\n", errno);
	return (-1);
    }
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = servers[busy].addr;
    if (servers[busy].pstate == PS_PORTM)
	addr.sin_port = htons(servers[busy].port-1);	/* try port-1 */
    else
	addr.sin_port = htons(servers[busy].port);	/* do normal port */
    if (connect(sock, &addr, sizeof(addr)) < 0) {
	if (errno != EINPROGRESS) {
#ifdef DEBUG
	    perror("connect (server connection)");
#endif
	    log_msg("connect() failed in open_server, err=%d", errno);
	    close(sock);
	    return (-1);
	} else {
	    /* connect has not taken effect yet */
	    FD_SET(sock, &writefd_set);
	    FD_CLR(sock, &readfd_set);
	    servers[busy].cstate = CS_CONNECTING;
	}
    } else {
	/* connect succeeded immediately */
	FD_SET(sock, &readfd_set);
	servers[busy].cstate = CS_CONNECTED;
    }

    busy_sock = sock;

    new_conn = TRUE;		/*tell server.c that this is a new connection*/
    mnew_conn = TRUE;

    return (0);
}


/*
 * Looks like the server is finally connected.  Add it to the readfd list.
 */
/*static*/ int
server_connected(fd)
int fd;
{
    int flags;

    if (servers[busy].cstate != CS_CONNECTING)
	fprintf(stderr, "Internal WARNING: cstate != CONNECTING\n");
    FD_CLR(fd, &writefd_set);
    FD_SET(fd, &readfd_set);
    servers[busy].cstate = CS_CONNECTED;

    if ((flags = fcntl(fd, F_GETFL, 0)) < 0) {
	perror("unable to get flags for NB clear (sc)");
	return;
    }
    if (fcntl(fd, F_SETFL, flags & (~O_NDELAY)) < 0) {
	perror("unable to set NB mode for connect (sc)");
	return;
    }

    /* note that the log message is also a "verbose" thang */
    if (verbose) {
	printf("%s: server connection established (%d)\n", prog, fd);
	log_msg("connected to %s (%d)", servers[busy].hostname,
		servers[busy].port -
		((servers[busy].pstate == PS_PORTM) ? 1 : 0));
    }
    return (0);
}


/*
 * Close connection to server.  Clears busy_sock.
 */
void
close_server()
{
    SERVER *sp;

    /* close the connection */
#ifdef DEBUG
    printf("Clearing busy_sock (%d)\n", busy_sock);
#endif
    close(busy_sock);
    FD_CLR(busy_sock, &readfd_set);
    FD_CLR(busy_sock, &writefd_set);
    busy_sock = -1;
    sp = &servers[busy];
    sp->cstate = CS_CLOSED;
}


/*
 * ---------------------------------------------------------------------------
 *	higher-level connection stuff
 * ---------------------------------------------------------------------------
 */

/*
 * Prepare a server connection.  Opens the first connection and sets up
 * the struct.
 */
int
prep_server()
{
    SERVER *sp;
    int i;

    sp = &servers[busy];

    /* hack to avoid the "(scanning server now)" message */
    bcopy(sp, &prev_info, sizeof(SERVER));

    sp->last_update = time(0);		/* in case we get a SS_NOCONN */

    /* clear out all the player info */
    sp->player_count = sp->queue_size = 0;
    for (i = 0; i < MAX_PLAYER; i++)
	sp->players[i].p_status = PFREE;

    sp->status = SS_WORKING;
    sp->why_dead = WD_UNKNOWN;

    if (try_alt_port)
	sp->pstate = PS_PORTM;
    else
	sp->pstate = PS_PORT;

    busy_time = time(0);
    if (open_server() < 0) {
	log_msg("open_server failed for %s", servers[busy].hostname);
	sp->status = SS_NOCONN;
	sp->why_dead = WD_CONN;
	return (-1);
    }

    return (0);
}


/*
 * Server stuff in progress.
 */
int
handle_server(fd)
{
    int cc;

    if (fd != busy_sock) {
	fprintf(stderr, "Internal ERROR: fd != busy sock in handle_server\n");
	fprintf(stderr, "fd = %d, busy_sock = %d\n", fd, busy_sock);
	log_msg(
	    "exit: Internal ERROR: fd != busy_sock in handle_server (%d, %d)\n",
	    fd, busy_sock);
	exit(1);
    }

    if (servers[busy].pstate == PS_PORTM) {
	if ((cc = read_minus(fd)) < 0) {
	    if (cc == -2) {
		/* not finished reading yet */
		return;
	    }
	    /* port-1 doesn't exist; move on to normal port */
#ifdef DEBUG
	    printf("No port-1 for this server\n");
#endif
	    close_server();
	    servers[busy].pstate = PS_PORT;
	    if (open_server() < 0) {
		log_msg("open_server failed (a)");
		servers[busy].status = SS_NOCONN;
		servers[busy].why_dead = WD_CONN;
		nuke_server();
	    }
	} else if (cc == SS_EMPTY) {
	    /* port exists, but server is empty */
	    servers[busy].status = SS_EMPTY;
	    close_server();
	    nuke_server();
	} else {
	    /* we got info back; go look for wait queue or better info */
	    close_server();
	    servers[busy].pstate = PS_PORT;
	    if (open_server() < 0) {
		/* weird... 2591 is up, but 2592 isn't */
		log_msg("open_server failed (b)");
		servers[busy].status = SS_NOCONN;
		servers[busy].why_dead = WD_CONN;
		nuke_server();
	    }
	}

    } else if (servers[busy].pstate == PS_PORT) {
	if ((cc = read_server(fd)) < 0) {
	    if (cc == -2) {
		/* not finished reading from server; keep going */
		return;
	    }
	    /* server must not exist */
#ifdef DEBUG
	    printf("Server dead\n");
#endif
	    servers[busy].status = SS_NOCONN;
	    /*servers[busy].why_dead = WD_CONN;*/	/* set by r_s() */
	    close_server();
	    nuke_server();
	} else {
	    /* success! */
#ifdef DEBUG
	    printf("Success: status %d\n", cc);
#endif
	    servers[busy].status = cc;
	    close_server();
	    nuke_server();
	}

    } else {
	fprintf(stderr, "Internal ERROR: bogus pstate in handle_server\n");
	log_msg("exit: Internal ERROR: bogus pstate in handle_server");
	exit(1);
    }
}


/*
 * Update the structures associated with the server connection.  Assumes
 * that the sockets have been closed.  The "status" field should be set.
 *
 * Clears "busy"; do not use it after calling this.
 *
 * This also sets various flags, so the "status" field should be set before
 * calling here.
 */
int
nuke_server()
{
    SERVER *sp;
    int plustime;

    sp = &servers[busy];

    switch (sp->status) {
    case SS_NOCONN:
	plustime = wait_noconn;
	break;
    case SS_EMPTY:
	plustime = wait_empty;
	break;
    case SS_OPEN:
	plustime = wait_open;
	break;
    case SS_QUEUE:
	plustime = wait_queue;
	break;
    default:
	fprintf(stderr, "Internal ERROR: strange 'why' in nuke_server\n");
	break;
    }

    /* update the struct */
    sp->last_update = time(0);
    sp->next_update = time(0) + plustime*60;

    /* tell it to pick another real soon */
    coffee = TRUE;

    /* while we're here, update the server's display_flags */
    if (sp->status == SS_QUEUE || sp->player_count >= 8)
	sp->display_flags |= DF_HAD_TMODE;
    if (sp->status != SS_NOCONN)
	sp->display_flags &= ~(DF_ALWAYS_DEAD);

    if (verbose) printf("%s: server conn closed (status %d); next at %s", prog,
	sp->status, ctime(&sp->next_update));

    busy = -1;
    return (0);
}


#endif

/*
 * Check the status of our server-scanning activities.  Also looks for
 * user connections which have been open too long.
 */
int
check_status()
{
    register USER *up;
    time_t now;
    register int i;
    static time_t last_user_check;		/* don't nuke the CPU */

    now = time(0);

    /* time out stale user connections */
    if (now - last_user_check > MAX_USER) {
	for (i = 0, up = users; i < maxfd; i++, up++) {
	    if (up->start_time && verbose) printf("active user %d\n", i);
	    if (up->start_time && (now - up->start_time >= MAX_USER)){
		close_user(i);
		if (verbose) printf("%s: hosed %d\n", prog, i);
	    }
	}

	last_user_check = now;
    }

#ifdef FUBAR
    /* check up on the server connection */
    now = time(0);
    if (busy >= 0) {
	/* see how long we've been busy (note busy_time can be > now) */
	if (now - busy_time > MAX_BUSY) {
	    /* we must've stalled... hose this one */
#ifdef DEBUG
	    printf("Server not responding\n");
#endif
	    if (verbose) log_msg("Out of time for %s", servers[busy].hostname);
	    servers[busy].status = SS_NOCONN;
	    servers[busy].why_dead = WD_TIMEOUT;
	    close_server();
	    nuke_server();
	    /* fall through and pick another */
	} else {
	    /* doing just fine */
#ifdef DEBUG
	    printf("Status checked\n");
#endif
	    return;
	}
    }

    if (busy >= 0) {
	fprintf(stderr, "Internal ERROR: still busy??\n");
	log_msg("exit: Internal ERROR: still busy?");
	exit(3);
    }
#endif
}


/*
 * Wait and watch for user connections.  If the user wants some out-of-date
 * info, fire up the MS-II connection.
 */
void
main_loop()
{
    register int i, j;
    struct sockaddr_in addr;
    int on = 1;
    fd_set readfds, writefds;
    int cc, lsock, *listen_sock = NULL;
    struct itimerval udt;

    /* init the fd set */
    FD_ZERO(&readfd_set);
    FD_ZERO(&writefd_set);

    /* Prepare sockets to accept calls */
    if ((listen_sock = (int *) malloc(port_count * sizeof(int))) == NULL) {
	fprintf(stderr, "ERROR: malloc failure (main/listen)\n");
	exit(1);
    }
    for (i = 0; i < port_count; i++) {
	if ((lsock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	    perror("ERROR: unable to create socket to listen to");
	    log_msg("exit: unable to create listen socket");
	    exit(1);
	}
	setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(ports[i].port);
	if (bind(lsock, &addr, sizeof(addr)) < 0) {
	    perror("ERROR: unable to bind to socket");
	    log_msg("exit: unable to bind to listen socket");
	    exit(1);
	}
	if (listen(lsock, 5) < 0) {
	    perror("ERROR: unable to listen to socket\n");
	    log_msg("exit: unable to listen to listen socket");
	    exit(1);
	}
	FD_SET(lsock, &readfd_set);

	listen_sock[i] = lsock;
	if (verbose) printf("%s: Listening on port %d\n", prog,
	    ntohs(addr.sin_port));
    }

    if (verbose) printf("--- Ready to accept connections\n");

    coffee = TRUE;	/* start right away */
    udt.it_interval.tv_sec = TIME_RES;
    udt.it_interval.tv_usec = 0;
    udt.it_value.tv_sec = TIME_RES;
    udt.it_value.tv_usec = 0;
    setitimer(ITIMER_REAL, &udt, 0);


    while (1) {
	if (coffee) {
	    /* got a wakeup signal, check status of server probe */
	    check_status();
	    coffee = FALSE;
	}
	bcopy(&readfd_set, &readfds, sizeof(fd_set));
	bcopy(&writefd_set, &writefds, sizeof(fd_set));
	if ((cc = select(maxfd, &readfds, &writefds, NULL, NULL)) <= 0) {
#ifdef DEBUG
	    if (cc < 0) {
		fprintf(stderr, "%s: ", prog);
		perror("select");
		/* usually an interrupted system call */
	    }
#endif
	    continue;
	}
	if (verbose) fseek(stdout, (off_t) 0, 2);

	/* cc is #of interesting file descriptors */
	for (i = 0; i < maxfd && cc; i++) {
	    if (FD_ISSET(i, &readfds)) {
		for (j = 0; j < port_count; j++) {
		    /* is it a user request? */
		    if (i == listen_sock[j]) {
			new_user(j, i);	/* port index, fd */
			break;
		    }
		}
		if (j == port_count) {
		    /* must be news from the metaserver */
		    /*handle_server(i);*/
		}
		cc--;
	    } else if (FD_ISSET(i, &writefds)) {
		if (i == busy_sock) {
		    /* finally got ahold of the darn metaserver */
		    /*server_connected(i);*/
		    /*handle_server(i);*/
		} else {
		    /* looks like a user is writeable */
		    handle_user(i);
		}
		cc--;
	    }

	}

	if (verbose) fflush(stdout);
    }

    /*NOTREACHED*/
}

